/*
 * Copyright (C) 2001-2006 by egnite Software GmbH
 * Copyright (C) 2010-2011 by egnite GmbH
 * Copyright (C) 2012 by Michael Fischer
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

#include <string.h>

#include <arch/nios2.h>
#include <dev/irqreg.h>

#include <cfg/uart.h>
#include <dev/debug.h>

#include <sys/device.h>
#include <sys/file.h>
#include <sys/timer.h>

/* Nios II */
#include "system.h"
#include "arch/nios2/nios_io.h"
#include "altera_avalon_uart.h"
#include "altera_avalon_uart_regs.h"

static NUTFILE dbgfile;

/*!
 * \brief Handle I/O controls for debug device.
 *
 * The debug device supports UART_SETSPEED only.
 *
 * \return 0 on success, -1 otherwise.
 */
int DebugIOCtl(NUTDEVICE *dev, int req, void *conf)
{
   uint32_t Divisor;
   uint32_t Baudrate = *((uint32_t*)conf);

   if (req == UART_SETSPEED)
   {
      Divisor = (((UART_0_FREQ*10) / Baudrate) + 5) / 10;
      IOWR_ALTERA_AVALON_UART_DIVISOR(dev->dev_base, Divisor);
      return(0);
   }

   return(-1);
}

/*!
 * \brief Initialize debug device.
 *
 * Simply enable the device, 115.2 kBaud.
 *
 * \return Always 0.
 */
int DebugInit(NUTDEVICE *dev)
{
   uint32_t Baud = 115200;
   
   DebugIOCtl(dev, UART_SETSPEED, &Baud); 
   
   return(0);
}

/*!
 * \brief Send a single character to debug device 0.
 *
 * A newline character will be automatically prepended
 * by a carriage return.
 */
static void DebugPut(CONST NUTDEVICE *dev, char ch)
{
   uint32_t status;

   if (ch == '\n') 
   {
      status = IORD_ALTERA_AVALON_UART_STATUS(dev->dev_base);
      while ((status & ALTERA_AVALON_UART_STATUS_TRDY_MSK) == 0)
      {
         status = IORD_ALTERA_AVALON_UART_STATUS(dev->dev_base);
      }   
      IOWR_ALTERA_AVALON_UART_TXDATA(dev->dev_base, '\r');   
   }
        
   status = IORD_ALTERA_AVALON_UART_STATUS(dev->dev_base);
   while ((status & ALTERA_AVALON_UART_STATUS_TRDY_MSK) == 0)
   {
      status = IORD_ALTERA_AVALON_UART_STATUS(dev->dev_base);
   }   
   IOWR_ALTERA_AVALON_UART_TXDATA(dev->dev_base, ch);   
}

/*!
 * \brief Send characters to debug device.
 *
 * A carriage return character will be automatically appended
 * to any linefeed.
 *
 * \return Number of characters sent.
 */
int DebugWrite(NUTFILE *fp, CONST void *buffer, int len)
{
   int c = len;
   CONST char *cp = buffer;

   while (c--) 
   {
      DebugPut(fp->nf_dev, *cp++);
   }
   
   return(len);
}

#ifdef NUT_DEV_DEBUG_READ
/*!
 * \brief Read characters from debug device.
 *
 * This function is called by the low level input routines of the
 * \ref xrCrtLowio "C runtime library", using the _NUTDEVICE::dev_read
 * entry.
 *
 * The function will block the calling thread until at least one
 * character has been received.
 *
 * \param fp     Pointer to a \ref _NUTFILE structure, obtained by a
 *               previous call to DebugOpen().
 * \param buffer Pointer to the buffer that receives the data. If zero,
 *               then all characters in the input buffer will be
 *               removed.
 * \param size   Maximum number of bytes to read.
 *
 * \return The number of bytes read, which may be less than the number
 *         of bytes specified. A return value of -1 indicates an error,
 *         while zero is returned in case of a timeout.
 */
int DebugRead(NUTFILE *fp, void *buffer, int size)
{
   int rc;
   unsigned int ch;
   char *bp = (char *) buffer;
   NUTDEVICE *dev = fp->nf_dev;

   /* Wait for the first character, forever. */
   for (rc = 0; rc < size; rc++) 
   {
      while ((IORD_ALTERA_AVALON_UART_STATUS(dev->dev_base) & ALTERA_AVALON_UART_STATUS_RRDY_MSK) == 0)
      {
         NutSleep(1);
         if ((rc || bp == NULL) && ((IORD_ALTERA_AVALON_UART_STATUS(dev->dev_base) & ALTERA_AVALON_UART_STATUS_RRDY_MSK) == 0))
         {
            return(rc);
         }
      }
      ch = IORD_ALTERA_AVALON_UART_RXDATA(dev->dev_base);
      if (bp) 
      {
         if (ch == '\r') 
         {
            *bp++ = '\n';
         } 
         else 
         {
            *bp++ = (char) ch;
         }
      }
   }
   return(rc);
}

/*!
 * \brief Retrieves the number of characters in input buffer.
 *
 * This function is called by the low level size routine of the C runtime
 * library, using the _NUTDEVICE::dev_size entry.
 *
 * \param fp     Pointer to a \ref _NUTFILE structure, obtained by a
 *               previous call to DebugOpen().
 *
 * \return The number of bytes currently stored in input buffer.
 */
long DebugSize(NUTFILE *fp)
{
   if (IORD_ALTERA_AVALON_UART_STATUS(fp->nf_dev->dev_base) & ALTERA_AVALON_UART_STATUS_RRDY_MSK)
   {
      return(1);
   }
   
   return(0);
}

#endif

/*!
 * \brief Open debug device.
 *
 * \return Pointer to a static NUTFILE structure.
 */
NUTFILE *DebugOpen(NUTDEVICE *dev, CONST char *name, int mode, int acc)
{
   dbgfile.nf_next = 0;
   dbgfile.nf_dev = dev;
   dbgfile.nf_fcb = 0;
  
   return(&dbgfile);
}

/*!
 * \brief Close debug device.
 *
 * \return Always 0.
 */
int DebugClose(NUTFILE *fp)
{
   return(0);
}

/*!
 * \brief Debug device 0 information structure.
 */
NUTDEVICE devDebug0 = 
{
   0,             /*!< Pointer to next device, dev_next. */
   {'u', 'a', 'r', 't', '0', 0, 0, 0, 0},    /*!< Unique device name, dev_name. */
   0,             /*!< Type of device, dev_type. */
   UART_0_BASE,   /*!< Base address, dev_base. */
   0,             /*!< First interrupt number, dev_irq. */
   0,             /*!< Interface control block, dev_icb. */
   0,             /*!< Driver control block, dev_dcb. */
   DebugInit,     /*!< Driver initialization routine, dev_init. */
   DebugIOCtl,    /*!< Driver specific control function, dev_ioctl. */
#ifdef NUT_DEV_DEBUG_READ
   DebugRead,     /*!< dev_read. */
#else   
   0,             
#endif   
   DebugWrite,    /*!< dev_write. */
   DebugOpen,     /*!< dev_opem. */
   DebugClose,    /*!< dev_close. */

#ifdef NUT_DEV_DEBUG_READ
   DebugSize      /*!< dev_size. */
#else
   0
#endif
};

/*** EOF ***/

