/*
 * Copyright (C) 2008-2009 by egnite GmbH
 * Copyright (C) 2012 by Michael Fischer
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

#include <sys/timer.h>
#include <sys/nutdebug.h>

#include <dev/spibus.h>
#include <arch/nios2/dev/spibus.h>
#include <dev/irqreg.h>
#include <sys/event.h>

/* NiosII */
#include "system.h"
#include "arch/nios2/nios_io.h"

#if   (SPI_0_IRQ == 0)
#define SPI_SIGNAL   sig_INTERRUPT0
#elif (SPI_0_IRQ == 1)
#define SPI_SIGNAL   sig_INTERRUPT1
#elif (SPI_0_IRQ == 2)
#define SPI_SIGNAL   sig_INTERRUPT2
#elif (SPI_0_IRQ == 3)
#define SPI_SIGNAL   sig_INTERRUPT3
#elif (SPI_0_IRQ == 4)
#define SPI_SIGNAL   sig_INTERRUPT4
#elif (SPI_0_IRQ == 5)
#define SPI_SIGNAL   sig_INTERRUPT5
#elif (SPI_0_IRQ == 6)
#define SPI_SIGNAL   sig_INTERRUPT6
#elif (SPI_0_IRQ == 7)
#define SPI_SIGNAL   sig_INTERRUPT7
#else
   Error: Wrong SPI_0_IRQ;
#endif


//#define SPIBUS0_POLLING_MODE


/*! 
 * \brief Select a device on the first SPI bus.
 *
 * Locks and activates the bus for the specified node.
 *
 * \param node Specifies the SPI bus node.
 * \param tmo  Timeout in milliseconds. To disable timeout, set this
 *             parameter to NUT_WAIT_INFINITE.
 *
 * \return 0 on success. In case of an error, -1 is returned and the bus 
 *         is not locked.
 */
static int Nios2SpiBus0Select (NUTSPINODE *node, uint32_t tmo)
{
   int rc = 0;
   
   (void)node;
   (void)tmo;
   
   return(rc);
}  

/*!
 * \brief Deselect a device on the first SPI bus.
 *
 * Deactivates the chip select and unlocks the bus.
 *
 * \param node Specifies the SPI bus node.
 *
 * \return Always 0.
 */
static int Nios2SpiBus0Deselect (NUTSPINODE *node)
{
   int rc = 0;
   
   (void)node;
   
   return(rc);
}

#if defined(SPIBUS0_POLLING_MODE)
/*! 
 * \brief Transfer data on the SPI bus in polling mode.
 *
 * A device must have been selected by calling Nios2SpiBus0Select().
 *
 * \param node Specifies the SPI bus node.
 * \param txbuf Pointer to the transmit buffer. If NULL, undetermined
 *              byte values are transmitted.
 * \param rxbuf Pointer to the receive buffer. If NULL, then incoming
 *              data is discarded.
 * \param xlen  Number of bytes to transfer.
 *
 * \return Always 0.
 */
static int Nios2SpiBus0PollTransfer (NUTSPINODE *node, CONST void *txbuf, void *rxbuf, int xlen)
{
   uint8_t b = 0xff;
   uint8_t *txp = (uint8_t *) txbuf;
   uint8_t *rxp = (uint8_t *) rxbuf;
   uintptr_t base;

   /* Sanity check. */
   NUTASSERT(node != NULL);
   NUTASSERT(node->node_bus != NULL);
   NUTASSERT(node->node_bus->bus_base != 0);
   base = node->node_bus->bus_base;

   while (xlen--) 
   {
      if (txp)
      {
         b = *txp++;
      }
      /* Transmission starts by writing the transmit data. */
      IOWR(base, SPI_TXDATA_OFF, b);
      
      /* Wait for receiver data register full. */
      while((IORD(base, SPI_STATUS_OFF) & SPI_STATUS_RRDY) == 0);

      /* Read incoming data. */
      b = (uint8_t)IORD(base, SPI_RXDATA_OFF);
      if (rxp) 
      {
         *rxp++ = b;
      }
   }
   return(0);
}
#endif /* #if defined(SPIBUS0_POLLING_MODE) */

#if !defined(SPIBUS0_POLLING_MODE)

static uint8_t * volatile spi0_txp;
static uint8_t * volatile spi0_rxp;
static volatile size_t spi0_xc;

void Nios2SpiBus0Interrupt(void *arg)
{
   uint8_t b;

   /* Get the received byte. */
   b = (uint8_t)IORD(SPI_0_BASE, SPI_RXDATA_OFF);
   if (spi0_xc) 
   {
      if (spi0_rxp) 
      {
         *spi0_rxp++ = b;
      }
      spi0_xc--;
   }
   
   if (spi0_xc) 
   {
      if (spi0_txp) 
      {
         b = *spi0_txp++;
      }
      IOWR(SPI_0_BASE, SPI_TXDATA_OFF, b);
   } 
   else 
   {
      NutEventPostFromIrq((void **)arg);
   }
}

/*! 
 * \brief Transfer data on the SPI bus using single buffered interrupt mode.
 *
 * A device must have been selected by calling At91SpiSelect().
 *
 * \param node  Specifies the SPI bus node.
 * \param txbuf Pointer to the transmit buffer. If NULL, undetermined
 *              byte values are transmitted.
 * \param rxbuf Pointer to the receive buffer. If NULL, then incoming
 *              data is discarded.
 * \param xlen  Number of bytes to transfer.
 *
 * \return Always 0.
 */
int Nios2SpiBus0Transfer(NUTSPINODE * node, CONST void *txbuf, void *rxbuf, int xlen)
{
   uint8_t b = 0xff;
   uintptr_t base;
   uint32_t control;

   /* Sanity check. */
   NUTASSERT(node != NULL);
   NUTASSERT(node->node_bus != NULL);
   NUTASSERT(node->node_bus->bus_base != 0);
   base = node->node_bus->bus_base;

   if (xlen) 
   {
      spi0_txp = (uint8_t *) txbuf;
      spi0_rxp = (uint8_t *) rxbuf;
      spi0_xc = (size_t) xlen;
      if (spi0_txp) 
      {
         b = *spi0_txp++;
      }
      
      /* Enable and kick interrupts. */
      control  = IORD(base, SPI_CONTROL_OFF);
      control |= (SPI_CONTROL_IE    |    /* Eenables interrupts for any error condition. */
                  SPI_CONTROL_IRRDY );   /* Enables interrupts for the receiver ready condition. */
      IOWR(base, SPI_CONTROL_OFF, control);

      /* Send first data */      
      IOWR(base, SPI_TXDATA_OFF, b);

      /* Wait until transfer has finished. */
      NutEventWait(&node->node_bus->bus_ready, NUT_WAIT_INFINITE);
      
      /* Disable interrupt in the controller */
      IOWR(base, SPI_CONTROL_OFF, 0);
   }
   return(0);
}
#endif /* #if !defined(SPIBUS0_POLLING_MODE) */

/*! 
 * \brief Initialize an SPI bus node. 
 *
 * This routine is called for each SPI node, which is registered via 
 * NutRegisterSpiDevice().
 *
 * \param node Specifies the SPI bus node.
 *
 * \return 0 on success or -1 if there is no valid chip select.
 */
static int Nios2SpiBus0NodeInit (NUTSPINODE *node)
{
   int rc = 0;
   NUTSPIBUS *bus;
   
   /* Sanity check. */
   NUTASSERT(node != NULL);
   NUTASSERT(node->node_bus != NULL);
   bus = node->node_bus;
   
   /*
    * Mode, Rate and Bits can not configured.
    */
   
#if !defined(SPIBUS0_POLLING_MODE)
   /* First disable interrupt in the controller */
   IOWR(bus->bus_base, SPI_CONTROL_OFF, 0);
   
   NutRegisterIrqHandler(bus->bus_sig, Nios2SpiBus0Interrupt, &bus->bus_ready);

   NutIrqEnable(bus->bus_sig);
#endif
   
   return(rc);
}

/*!
* \brief Avr32 SPI bus driver implementation structure.
*/
NUTSPIBUS spiBus0Nios2 =
{
   NULL,                      /*!< Bus mutex semaphore (bus_mutex). */
   NULL,                      /*!< Bus ready signal (bus_ready). */
   SPI_0_BASE,                /*!< Bus base address (bus_base). */
   &SPI_SIGNAL,               /*!< Bus interrupt handler (bus_sig). */
   
   Nios2SpiBus0NodeInit,      /*!< Initialize the bus (bus_initnode). */
   Nios2SpiBus0Select,        /*!< Select the specified device (bus_alloc). */
   Nios2SpiBus0Deselect,      /*!< Deselect the specified device (bus_release). */
   
#if defined(SPIBUS0_POLLING_MODE)
   Nios2SpiBus0PollTransfer,  /*!< Transfer data to and from a specified node (bus_transfer). */
#else
   Nios2SpiBus0Transfer,
#endif   
   
   NutSpiBusWait,             /*!< Wait for bus transfer ready (bus_wait). */
   NutSpiBusSetMode,          /*!< Set SPI mode of a specified device (bus_set_mode). */
   NutSpiBusSetRate,          /*!< Set clock rate of a specified device (bus_set_rate). */
   NutSpiBusSetBits           /*!< Set number of data bits of a specified device (bus_set_bits). */
};

/*** EOF ***/
